/** ===========================================================================
	* Config
	* ======
	* Send this into the constructor, instead of long and complex argument lists.
	* We thus extend the JSX API without polluting the JOS API.
	* Use the same for both ObjOut (serialization) and ObjIn (deserialization).
	* Less elegant, and redundant; but simpler to write, maintain and use.
	* 
	* C++ style for setters enables this usage:
	*  Config c = new Config().aliasesUsed(true).IDsUsed(true);
	*
	* Nice: to design to serialize to XML nicely.  For example:
	* Sub-sets of functionality would have their own object.
	*  Config c = (Config) JSX.ObjIn(new FileReader("cyclicConfig.jsx"));
	**/

package JSX;

public class Config {
	public Config() {}	//because default visibily is package

//ObjIn - fields missing from classes
//Handling NoSuchFieldExceptions:
	// This handles fields in XML which are not found in the class.
	//CONSIDER: a better (read: less verbose) name for this?
	public NoSuchFieldHandler noSuchFieldHandler;
	public Config setNoSuchFieldHandler(NoSuchFieldHandler noSuchFieldHandler) {
		this.noSuchFieldHandler = noSuchFieldHandler;
		return this;
	}

//ObjOut - formatting
//Gil Peeter's compact "no formatting" (no unnec whitespace, all one line)
  public boolean formatted = true;	//easy test
	public Config setFormatted(boolean formatted) {
		this.formatted = formatted;
		return this;
	}

//ObjOut - superVersion
	public boolean superVersion;
	public Config superVersion(boolean superVersion) {
		this.superVersion = superVersion;
		return this;
	}

//ObjOut - aliases
/*
	public boolean areAliasesUsed;
	public Config aliasesUsed(boolean areAliasesUsed) {
		this.areAliasesUsed = areAliasesUsed;
		return this;
	}
*/

//ObjOut - ID's
/*
	public boolean areIDsUsed;
	public Config IDsUsed(boolean areIDsUsed) {
		this.areIDsUsed = areIDsUsed;
		return this;
	}
*/
}
