/** (c) Brendan Macmillan 2001, licensed under GNU's GPL 2 
    (Also available for non-GPL, commercial use, for fee) 
    Website: http://www.csse.monash.edu.au/~bren/JSX
    Website: http://freshmeat.net/projects/jsx
    List (can read without subscribing, need to join to post):
      http://groups.yahoo.com/group/JSX-ideas/messages
		Commercial licensing enquiries only:
			bren@mail.csse.monash.edu.au */


/** ===========================================================================
	* ParseUtilities
	* ==============
	* boolean, float and double; and
	* XML Strings (both parsing and writing code)
	*
	* Correctly parses NaN, Infinity and -Infinity of floating point numbers
	* (float and double).  I wonder why the standard versions do not do this
	* already?
	*
	* Based on a code patch submitted by Mark Collette, 11 Jul 2001
	**/
package JSX;

class ParseUtilities {
	public static boolean parseBoolean(String in) {
		String s = in.trim();
		if (s.equalsIgnoreCase("false")) return false;
		else if (s.equalsIgnoreCase("true")) return true;
		else
			throw new IllegalArgumentException("Boolean must be 'true' or 'false'; not '"
				+s+"')");
			//superclass of NumberFormatException
	}
	public static float parseFloat(String in) {
		String s = in.trim();
		if( s.equals("NaN") ) return Float.NaN;
		else if( s.equals("Infinity") ) return Float.POSITIVE_INFINITY;
		else if( s.equals("-Infinity") ) return Float.NEGATIVE_INFINITY;
		else return Float.parseFloat(s);
	}

	public static double parseDouble(String s) {
		if( s.equals("NaN") ) return Double.NaN;
		else if( s.equals("Infinity") ) return Double.POSITIVE_INFINITY;
		else if( s.equals("-Infinity") ) return Double.NEGATIVE_INFINITY;
		else return Double.parseDouble(s);
	}

	/** 
		* Changed to StringBuffer based on Raimi's "non-fork" version.
		* (FIXED) BUG: characters from (char)128 (0x80) to (char)159 (0x9f)
		* are encoded as "?"
		*
		* A new scheme:
		*   output "\_" as "\_"; unless _ is a \ or u
		*   (note: if the _ will need to be escaped, then it is a "\"...)
		* requires looking ahead to the next character.  That is, a digraph
		* encoding.
		* What to do if there is no next character?  I guess this could be
		* treated as a special digraph all its own, and given a particular
		* encoding.  It would be easiest for the decoder if this was encoded as
		* "\\".  That way, there is *always* another character following
		*/
	public static String encodeXML(String in) {
		StringBuffer out = new StringBuffer(in.length());	//cap at least this long
	//FIXME: \ u encoding <-- compiler picks it up as a malformed unicoding!
	//is \u0000 a legitimate char? Can it be rep in XML?
		for (int i=0; i<in.length(); i++) {
			char c = in.charAt(i);
			switch (c) {
				case '&': out.append("&amp;"); break;
				case '<': out.append("&lt;"); break;
				case '>': out.append("&gt;"); break;
				case '\'': out.append("&apos;"); break;
				case '"': out.append("&quot;"); break;
			//	case '\\': out.append("\\\\"); break;
				case '\\':
					//need to check for the next character, and encode them together:
					// if there is none, then output "\\" - need to escape for compiler
					// if it is '\', then output "\\\\"	- need to escape for compiler
					// if it is 'u', then output "\ \ u" (spaced so compiler not upset)
					// for all other cases, simply output '\' and the character.
					if (i+1==in.length())	{ //could also write as an Exception handler
						out.append("\\\\");	//escaping: two back-slashes \ \ 
						i++;	//we just encoded the next one
					}
					else {
						c = in.charAt(i+1);
						if (c=='\\') {
							out.append("\\\\\\\\");	//escaping: four back-slashes \ \ \ \ 
							i++;	//we just encoded the next one
						} else if (c=='u') {
							out.append("\\\\u");	//escaping: two back-slashes \ \ 
							i++;	//we just encoded the next one
						} else if (c<=0x1F || (c>=0x80 && c<=0x9f)) {
							out.append("\\\\");	//it *will* be followed by a \u0000
						} else {
							out.append("\\");	//leave the next char to be encoded as usual
						}
					}
					break;	//escape the escape char...
				case '\t': case '\n': case '\r': out.append(c); break;
				default:	//NB: exceptions done already (0x9 \t, 0xA \n, 0xD \r)
					//TODO: What about other ranges??
					//if (c<=0x1F || (c>=0x80 && c<=0x9f)) {	//char is unsigned, so no problem with -ve rep
					if (c<=0x1F || (c>=0x80)) {	//*all* non-ascii
 						out.append( "\\u" + Integer.toHexString(c+0x10000).substring(1) );
							//must be exactly 4 char long: add leading 0's, remove "1"
					} else {
 						out.append(c);
					}
			}
		}
		return out.toString();	//fast: a ref, not copy (unless needed).
	}

	/**--------------------------------------------------------------------------
		* decodeXML
		* ---------
		* &amp; etc.
		* To do: this also needs to be applied to char (not just Strings)
		* should rewrite to use StringBuffer - it presently creates new objects
		* like mad - a very expensive operation.
		**/
	public static String decodeXML(String in) {
		String a = new String("");	//prevent reference sharing for zero length
		int i = 0;
		try {
			for (i=0; i<in.length(); i++) {
				char c = in.charAt(i);
				if (c=='&') {				//if the escape character
					i++;
					if (in.startsWith("amp;", i))       {a+='&';  i+=3;} //NB: loop's i++
					else if (in.startsWith("lt;", i))   {a+='<';  i+=2;}
					else if (in.startsWith("gt;", i))   {a+='>';  i+=2;}
					else if (in.startsWith("apos;", i)) {a+='\''; i+=4;}
					else if (in.startsWith("quot;", i)) {a+='"';  i+=4;}
				}
				else if (c=='\\') {	//if \ escape char. NB: \ is quoted as \\ in source
					i++;
					if (in.charAt(i)=='u') { //NB: loops's i++ if \ u (\ then u)
						i++;
						a+=(char)Integer.parseInt(in.substring(i,i+4),16);
 						i+=3;
					}
					else if (in.charAt(i)=='\\') {
						 a+='\\'; i+=0;	//a double \ \ is now a \ \!
					}	//if \\ (\ then \)
					else	{ //completely use up the codespace!
						 a+='\\'; i-=1;	//unread the last char
					}
				}
				else {
					a += c;
				}
			}
		} catch (StringIndexOutOfBoundsException e) {
			System.err.println( "Seems to be an incomplete escaped entity or \\unnnn: "
				+in.substring(i) );
		} catch (IndexOutOfBoundsException e) {
			System.err.println( "Seems to be an incomplete escaped entity: "
				+in.substring(in.lastIndexOf('&')) );
		} catch (NumberFormatException e) {
			System.err.println( "Seems to be a faulty \\unnnn escaped control char: "
				+in.substring(i) );
		}
		return a;
	}

	static class Dollar {
		public static void main(String a[]) {
			if (a.length==0)
				a = new String[] {"$a$a$"};
			for (int i=0; i<a.length; i++) {
				String t;
				System.err.println(a[i] + " -> "
					+ (t=escapeDollar(a[i])) + " -> "
					+ descapeDollar(t) );
			}
		}
	}


	/**--------------------------------------------------------------------------
		* escapeDollar
		* ------------
		* This converts "$" to DOLLAR (eg "DOLLAR-").  $ is legal in Java, but
		* not in XML.
		* Specifically, $ is legal *anywhere* in Java package, class and field
		* names (at beginning and at end - not just for inner classes!), but not
		* in XML tags and attributes.

		* The "-" is 100% illegal in Java package, class and field names, so is
		* safe.  To avoid conflict with "ArrayOf-", we prefix it to be "DOLLAR-".

		* NB: converting "$" to ".." doesn't work: both "a.$a" and "a$.a" map to
		* "a...a", so the decoding is ambiguous.
		*/
	//static final String DOLLAR = "-DOLLAR-";
	//static final String DOLLAR = "D-";
	//static final String DOLLAR = ".-";	//as released in JSX0.8.20.2
		//PROBLEM: leading $ (as in "class $SomeClass {}") is mapped to a leading
		//"." (ie <.-SomeClass), which is not well-formed XML.  We change it to
		//"_-" instead: this marks the unusal mapping fairly clearly.
	static final String DOLLAR = "_-";
	//It is very important that it is trailing a "-", so it cannot be
	//confused with "ArrayOf-".  Consider if "-DOLLAR" and "ArrayOf-DOLLAR"?
	//a fixed parsing order would disambiguate; but better not to need to do so.
	public static String escapeDollar(String in) {
		StringBuffer out = new StringBuffer();
		for (int i=0; i<in.length(); i++)	{ 	//step through
			char c;
			if ((c=in.charAt(i))=='$')
				out.append(DOLLAR);
			else
				out.append(c);
		}
		return out.toString();
	}

	/**--------------------------------------------------------------------------
		* descapeDollar  (see escapeDollar above for a fuller explanation)
		* -------------
		* This converts DOLLAR (eg "DOLLAR-") to "$".
		* One implementation will scan the input String with substring, for
		* DOLLAR-; from a marker, it will copy the previous stuff across, except
		* "$" instead of the "DOLLAR".  Then, update the marker.
		*/
	public static String descapeDollar(String in) {
		StringBuffer out = new StringBuffer();
		int marker = 0;
		while (true) {
			int newmarker = in.indexOf(DOLLAR, marker);
			if (newmarker==-1)
				break;	//no DOLLARs left
			out.append(in.substring(marker, newmarker)+'$');
			marker = newmarker+DOLLAR.length();
		}
		out.append(in.substring(marker));	//the rest of it (could be 0 length)
		return out.toString();
	}



  /** test class:
    * java JSX.ParseUtilities$Hex
    **/
	static class Hex {
		public static void main(String a[]) {
			if (a.length>0) {
				byte[] b = decodeHex(a[0]);
				System.err.print( print(b) );
				System.err.print( encodeHex(b) );
			}
			else {
				byte[] b = new byte[0x200];
				byte j=0;
  			for (int i=0; i<b.length; i++) b[i] = j++;	//j overflows silently at 127++
				String t;
				System.err.print(print(b) + " -> ");
				System.err.print((t=encodeHex(b)) + " -> ");
				System.err.print( print(decodeHex(t)) );
				System.err.println();
			}
		}
	}
	public static String print(byte[] in) {
		String out = "";
  	for (int i=0; i<in.length; i++) {
			out += in[i]+", ";		//ineff, but simple
		}
		return out;
	}

	/** one byte -> two chars.
		* TODO: need to make the \n platform independent:
		* 	out.append( '\n' );
		*
		* Some users may want to switch off the pretty printing, for a more compact
		* rendering.
		* A similar switch may be appreciated for indentation in general - it
		* can take up a lot of room (Gary M turned it off; Mark C noted it for
		* binary arrays that were nested deeply.
		**/
	public static String encodeHex(byte[] in) {
	   return encodeHex(in, true);
	}
	/** one byte -> two chars.
		* TODO: need to make the \n platform independent:
		* 	out.append( '\n' );
		*
		* Some users may want to switch off the pretty printing, for a more compact
		* rendering.
		* A similar switch may be appreciated for indentation in general - it
		* can take up a lot of room (Gary M turned it off; Mark C noted it for
		* binary arrays that were nested deeply.
		**/
	public static String encodeHex(byte[] in, boolean formatted) {
  	StringBuffer out = new StringBuffer(in.length*3);	//was 2
 		if (formatted)
 		  out.append( EOL_SEP );
  	for (int i=0; i<in.length;) {	//NB: i++ is defered
    	out.append( toHexChar((in[i]&0xF0)>>>4) );  //hi 4 bits, shifted
    	out.append( toHexChar((in[i]&0x0F)>>>0) );  //low 4 bits
			i++;
   		if (formatted)
   		{
			  if (i%256==0)
      		out.append( EOL_SEP+EOL_SEP );
			  else if (i%16==0)
      		out.append( EOL_SEP );
			  else if (i%4==0)
      		out.append( "  " );
			  else
      		out.append( ' ' );
      }
				//I think we should have no spacing at all here (?)
      //else
      //	out.append( ' ' );
  	}
 		if (formatted)
 		  out.append( EOL_SEP );	//make it platform independent
  	return out.toString();
	}
	static final String EOL_SEP = System.getProperty("line.separator");
 		//platform independence

	static char toHexChar(int b) {
		if (b>=10)
			return (char)(b-10+'A');
		else
			return (char)(b+'0');
	}


	/** 
		* decodeHex
		* two chars -> one byte
  	* no error checking for malformed hex encoding.
		* - if 0xF0 bits are ever set
		* - if an odd number of chars
		* Skipping ws is tricky, because we can't use the String length to 
		* estimate it.
  	**/
	public static byte[] decodeHex(String in) {
  	byte[] out = new byte[in.length()/2];
		int j = 0;	//need j's value after the loop
  	for (int i=0; i<in.length();) {
			char a = in.charAt(i++);
			if (Character.isWhitespace(a))
				continue;
    	out[j++] = (byte)
      	( ((fromHexChar(a) &0x0F)<<4) |
				  (( fromHexChar(in.charAt(i++)) &0x0F)<<0) ); //redundant for "human clarity"
			//Integer.parseInt(hexString, 16);   //use instead?
  	}
		if ( j<(in.length()/2) ) {
			byte[] trimOut = new byte[j];
			System.arraycopy(out, 0, trimOut, 0, j);
			out = trimOut;
		}
  	return out;
	}


	/**
		* Handles both upper and lower case.
		* Does not check that within range.
		**/
	static int fromHexChar(char a) {
		if (Character.isUpperCase(a))
			return (char)(a-'A'+10);
		else if (Character.isLowerCase(a))
			return (char)(a-'a'+10);
		else if (Character.isDigit(a))
			return (char)(a-'0');
		else throw new NumberFormatException("'"+a+"' is not a valid hex character");
	}

}
