/** (c) Brendan Macmillan 2001, licensed under GNU's GPL 2 
    (Also available for non-GPL, commercial use, for fee) 
    Website: http://www.csse.monash.edu.au/~bren/JSX
    Website: http://freshmeat.net/projects/jsx
    List (can read without subscribing, need to join to post):
      http://groups.yahoo.com/group/JSX-ideas/messages
		Commercial licensing enquiries only:
			bren@mail.csse.monash.edu.au */


/** ===========================================================================
	* ParserXML (JSX0.8).
	* =========
	*
 	* email comments to:
    bren@mail.csse.monash.edu.au (G. Brendan Macmillan)
 **/

/** Syntactic elements: <, =, "", > and /		(see end for further info)
	* DOCUMENT: ELEMENT*
	* ELEMENT: <ZZZZ ATTR*>ELEMENT*</ZZZZ>
	* ATTR: XXXX="YYYY"
	**/

/** PROBLEM: \n in quoted values.
	* SOLUTION: 
	*	 	(a). reset the syntax (so that \n \r are significant, not whitespace)
	*		(b). grab up to the next " char, and return as String
	*		(c). then reset syntax again.
	**/
	

package JSX;
import java.io.*;

public class ParserXML {
	Reader in;	//keep it so we can close later on...
	StreamTokenizer st;
	final static boolean DEBUG = false;
	//final static boolean DEBUG = true;

	final static boolean CLOSETAG_DEBUG = false;
	//final static boolean CLOSETAG_DEBUG = true;

	public ParserXML(Reader in) {
		this.in = in;
		st = new StreamTokenizer(new BufferedReader(in));	//thanks to Levi Mota
		setSyntax();
	}

	public void close() {
		try {
			in.close();	//there is no Reader.flush()
		} catch (IOException e) {
			e.printStackTrace();
		}
	}


	/**--------------------------------------------------------------------------
		* setSyntax
		* ---------
		*  Note: is altered in readAttr(), to cope with newlines within quotes.
		**/
	private void setSyntax() {
		st.eolIsSignificant(false);		//quoted text is broken by /n still
																	//checked src: yes, \n is hard-coded
		st.resetSyntax();	//(so numbers aren't significant)
		st.wordChars(0,255);
		st.ordinaryChar('<');		//could loop over a list...
		st.ordinaryChar('/');
		//st.ordinaryChar('?');	//for processing instruction PI
		st.ordinaryChar('>');
		st.ordinaryChar('=');
		st.ordinaryChar('"');	//detect for as a token for OWN IMPLEMENTATION
		st.ordinaryChar('\'');	//detect for as a token for OWN IMPLEMENTATION
		//st.quoteChar('"');
		st.whitespaceChars(' ',' ');
		st.whitespaceChars('\t','\t');
		st.whitespaceChars('\n','\n');
		st.whitespaceChars('\r','\r');
	}



	/**--------------------------------------------------------------------------
		* readAttr
		* --------
		* XXXX = "yyyy"  |  > 
		*
		**/
	public Attr readAttr() throws ExceptionXML, IOException {

		Attr attr = new Attr();
		int ttype;
		ttype = st.nextToken();		//expect an attribute name
		if (ttype=='>') {
			attr.isEnd = true;
			return attr;
		}

		if (ttype=='/') {	//first char of digraph for "EMPTY TAG"
			if (st.nextToken() == '>')		//expect an attribute name
			{
				attr.isEnd=true;
				attr.emptyTag=true;
				return attr;
			}
			else
			 throw new ExceptionXML("empty tag (\"/>\")", st);
		}

		if (ttype!=st.TT_WORD)
			throw new ExceptionXML("an attribute name", st);

		//We can do "?>" here -  (it is not same as ordinary char)
		if (st.sval.equals("?")) {	//first char of digraph for "END PI" is '?'
			if (st.nextToken() == '>')		//expect an attribute name
			{
				attr.isEndPI=true;	//special end logic, to prevent <tag ?>
				return attr;
			}
			else
			 throw new ExceptionXML("PI end (\"?>\")", st);
		}

		attr.name = st.sval;
		ttype = st.nextToken();		//expect a =
		if (ttype!='=')
			throw new ExceptionXML(
				"Expected \"=\" to follow attribute name \""  +attr.name+ "\""
			);
			//NOTE: Any ws between '=' and '"' has already been gobbled up; therefore,
			//it is OK to switch the syntax *now* - or else ws *following* the
			//'"' will also be snaffled.
			//OWN IMPLEMENTATION:
			  st.wordChars('>','>');	//by XML1.0, this is legal in an attr value
			  st.wordChars('=','=');
			  st.wordChars('/','/');
			  //st.wordChars('?','?');
			  st.ordinaryChar(' ');	//means it is returned individually, I think?
			  //st.wordChars(' ',' ');
			  st.ordinaryChar('\t');
			  //st.wordChars('\t','\t');
		st.eolIsSignificant(true);
			  //st.wordChars('\n','\n');	//these appear to be just skipped?
			  st.ordinaryChar('\r');
			  //st.wordChars('\r','\r');

		ttype = st.nextToken();		//expect a quoted value: OWN IMPLEMENTATION
//System.err.println("1 ttype of thing read: "+ ttype + ", " + (char)ttype);
		if (ttype!='"' && ttype!='\'')
			throw new ExceptionXML(
				"'quoted' or \"quoted\" value (single or double) to follow \""+
				attr.name + "=\"", st);		//"name="
		if (DEBUG) System.err.println(st.sval + " (sval) *with* quote");
		int quoteChar = ttype;
			 /* PROBLEM: a \n breaks a quoted string, finishing it.  I can't see how
				* to fix this.  You want it to ignore the \n within a quoted string.
				* other ws within a quoted string does not separate it into separate
				* tokens; but \n does, even when specified as not significant.  This
				* is a bug in the tokenizer; or else I am initializing it wrongly.
				*
				* PROBLEM: leading whitespace within a quote is ignored...
				* Cause: when first quote read, following whitespace is not
				* a word char and so is gobbled up.  Solution is to set syntax
				* earlier, as a quoted string *must* follow, or else error, anyway.
				*/

				ttype = st.nextToken();		//expect a quoted value
				if (ttype==quoteChar)	{  //ie empty string is possible! Romain bug
					attr.value = "";	//NB: not the same as null
				}
				else {
					if (DEBUG) System.err.println(ttype + " (type) read after quote");
					if (DEBUG) System.err.println(st.sval + " (sval) read after quote");
					//build up value in StringBuffer, for efficiency:
					StringBuffer out = new StringBuffer();	//don't know length...
						//assign String to attr.value when fully read...
					while (true) {	//read quoted value	HACK, but works. Tokenizer sucks
						//System.err.println("2 ttype of thing read: "+ ttype + ", " + (char)ttype);
						if (DEBUG) System.err.println("BUILDING: '"+out+"'");
						if (ttype==' ') { 	//eat ws
							out.append(' ');
								//attr.value += ' ';
					 		ttype = st.nextToken();
							continue;
						}
						if (ttype=='\t') { 	//eat ws
							out.append('\t');
								//attr.value += '\t';
					 		ttype = st.nextToken();
							continue;
						}
						if (ttype=='\r') { 	//eat ws
							out.append('\r');
								//attr.value += '\r';
					 		ttype = st.nextToken();
							continue;
						}
						if (ttype==st.TT_EOL) { 	//eat nl
							out.append('\n');
								//attr.value += '\n';
					 		ttype = st.nextToken();
							continue;
						}
						if (ttype==st.TT_WORD) {
							out.append(st.sval);//NB: any leading ws + store quoted value.
								//attr.value += st.sval;//NB: any leading ws + store quoted value.
					 		ttype = st.nextToken();
							continue;
						}
						break;	//unreachable code
/*
						throw new ExceptionXML(
							"a \"quoted\" value (same as opening quote) to follow \""+
							attr.name + "=\"", st);		//"name="
*/
					}
					attr.value = out.toString();	//assign at end of while
				}
				if (DEBUG) System.err.println(st.sval + " was value read after quote");
				if (DEBUG) System.err.println(attr.value + " was total value");
				if (DEBUG) System.err.println(ttype + " read after that");
				if (ttype!=quoteChar)		//the same, whatever it was
					throw new ExceptionXML(
					"matching '"+(char)quoteChar+"'", st);
			//switch syntax back (duplicating some code for eff).  It hurts me to dup
		st.eolIsSignificant(false);
			  st.ordinaryChar('>');
			  st.ordinaryChar('=');
			  st.ordinaryChar('/');
			  //st.ordinaryChar('?');
				st.whitespaceChars(' ',' ');
				st.whitespaceChars('\t','\t');
				st.whitespaceChars('\n','\n');
				st.whitespaceChars('\r','\r');
//		setSyntax();
		attr.name = ParseUtilities.descapeDollar(attr.name);
		return attr;
	}

	/**--------------------------------------------------------------------------
		* Attr class.	(either XXXX="yyyy" -or- ">")
		* ==========
		*		attribute name | >
		*		=:
		*		": expects a quoted value
		*
		*	return attribute OR null (if no attributes left)
		**/
	public class Attr implements Serializable{
		public String name;
		public String value;
		boolean nameMissing;	//only for Deserialization convenience.  Yuk!
		int length;	//HACK!!  for use in message passing.
		boolean isEndPI = false;	//distinguish "?>" from "/>"
		public boolean isEnd = false;	//usually it is.  We flag the end with this.
		public boolean emptyTag = false;
		public String toString() {
		/*
			StringWriter sw = new StringWriter();
			XMLSerialize.serialize(this, new PrintWriter(sw, true));//this works!
			return sw.toString();
		*/
			return name +"=\""+value+"\" , "
			+ ((emptyTag)?"empty":"")+((isEnd)?"end":"")+" tag"
			+ " nameMissing="+nameMissing;
		}
	}



	/**--------------------------------------------------------------------------
		* readTag
		* -------
		*
		**/
	public Tag readTag() throws ExceptionXML, IOException {
		Tag tag = new Tag();	//may be more efficient to reuse?
		int ttype;
		ttype = st.nextToken();		//expect an <
		if (ttype==st.TT_EOF)
			return null;
		if (ttype!='<')
				throw new ExceptionXML("an open tag", st);
		ttype = st.nextToken();		//expect either a name, or a /
		switch (ttype) {
			case StreamTokenizer.TT_WORD:	//must be constant
				tag.name = st.sval;
				tag.start = true;
				break;
			case '/':		//close tag: </tag>, not empty tag: <tag/>
				ttype = st.nextToken();		//expect a name to follow
				if (ttype!=st.TT_WORD)
					throw new ExceptionXML("a tag", st);
				tag.name = st.sval;
				tag.start = false;
				ttype = st.nextToken();		//expect a >
				if (ttype!='>')
					throw new ExceptionXML("a >", st);
				break;
			default:
				throw new ExceptionXML("either \"<tag\" or \"</tag\"", st);
		}
		if (tag.name.startsWith("?")) {
			tag.isPI = true;
			tag.name = tag.name.substring(1);	//strip the "?" for sake of ease
		}
		if (CLOSETAG_DEBUG) {
			System.err.println("RETURNING TAG: "+tag);
			try {
				throw new Exception();
			} catch (Exception e) {
				e.printStackTrace();
			}
		}
		return tag;
	}
	/**--------------------------------------------------------------------------
		* Tag class.
		* ==========
		* //empty tag: <tag/> not implemented
		* Should we include information here, about whether it is a PI or not?
		* ie if it begins with a "?", as in <?xml version="1.0"?> or
		* <?jsx version="0.8.20.2"?>.  Note that it ends with "?>", which is a
		* little like an empty tag "/>".
		**/
	public class Tag {
		public String name;
		boolean isPI;	//for <?target xxx="yyy"?> processing instruction (PI)
		public boolean start;	//true: <tag>, false: </tag>
		public String toString() { return "<"+(isPI?"?":(start?"":"/")) + name; }
	}




	/**--------------------------------------------------------------------------
		* Exceptions
		* ==========
		*	this subclassing of exceptions is *very* reuseful modularity!
		* probably should always send the st, so can output line number
	 	* (and possibly position, last token etc, too).
		* Just subclass IOException since that is what it really is, anyway
		**/
	static public class ExceptionXML extends IOException {
		public ExceptionXML(String expected) {
			super(expected);
		}
		public ExceptionXML(String expected, String got) {
			super("Expected \"" +expected+ "\""
				+ ", but got \"" +got+ "\"");
		}
		public ExceptionXML(String comment, String expected, String got) {
			super(comment+":  Expected \"" +expected+ "\""
				+ ", but got \"" +got+ "\"");
		}
		public ExceptionXML(String expected, StreamTokenizer st) {
			super(
				"Line " +st.lineno()+": "
				+ "Expected " +expected
				+ ", but got "
					+"'"+
						(st.ttype!=StreamTokenizer.TT_WORD?
						(""+(char)(st.ttype)):
						(st.sval))
				  +"'     "
					+"["+st.ttype+": " +st.sval+ "]"
				);
		}
	}



	/**--------------------------------------------------------------------------
		* Test code
		* =========
		* Tag, fields, then tags again
		**/
	public void parse() {
		Tag tag;
		Attr attr;
		String indent = "";
		System.out.println(
			"Note: this version of parse() does no start-end tag balancing\n"+
			"Note: it does no checking for legality of class and field names\n");
		try {
			while ( (tag=readTag()) != null) {
				if (tag.start) {
					System.out.print(indent + tag);
					indent += "    ";             //whack on 4 spaces
					while ( (attr=readAttr()) != null)
						System.out.print(" "+attr);
				}
				else {
					indent = indent.substring(4);	//whack off 4 spaces
					System.out.print(indent + tag);
				}
				System.out.println(">");
			}
		}
		catch (Exception e) {
			System.out.println(e);
		}
	}


	public void dumbParse() {
		int ttype;
		try {
			while ( (ttype = st.nextToken()) != st.TT_EOF) {
				String a;
				switch (ttype) {
					case '"':
						a = "QUOTED:  "  +st.sval;
						break;
					case StreamTokenizer.TT_WORD:	//must be constant
						a = "TOKEN:  "  +st.sval;
						break;
					default:
						a = (char)ttype + " : " + st.sval;
						break;
				}
				System.out.println(a);
			}
		}
		catch (Exception e) {
			System.out.println(e);
		}
	}

	public static class Test {
		public static void main(String args[]) {
			new ParserXML(new InputStreamReader(System.in)).parse();
		}
	}
}

	/**--------------------------------------------------------------------------
		* Rough summary of syntax
		* -----------------------
		* ELEMENTS:
		*		<: OPEN_TAG
		*		</: CLOSE_TAG
		*		ELEMENTS
		* 
		* OPEN_TAG: expects a tag name
		*		FIELDS
		*		return
		*
		*	CLOSE_TAG: expects a tag name (AT SAME LEVEL-it can only be one, I think)
		*  >: return
		*
		* FIELDS:
		* 	>: return
		*		FIELD_NAME
		*		FIELDS
		**/
