
/*
 * Copyright (C) 2000-2001 Technical University of Ilmenau, Germany
 *                         Department of Microelectronic Circuits and Systems
 *
 * You may only use this file according to the terms of the license agreement included
 * within the file LICENSE which is distributed together with the software package.
 *
 * CVS Info:
 * $Author: rummler $
 * $Date: 2001/11/13 09:50:33 $
 * $Revision: 1.5 $
 * $State: Exp $
 * $RCSfile: ListMutationExample.java,v $
 *
 */

package mss.ea.ex.mutation;

import java.util.*;

import mss.ea.al.*;
import mss.ea.core.*;
import mss.ea.eval.*;
import mss.ea.ins.*;
import mss.ea.mut.*;
import mss.ea.rec.*;
import mss.ea.stat.*;
import mss.ea.trm.*;

/** Simple example of how to mutate lists.
 * 
 * @since 0.2
 * @author Andreas Rummler
 * @version $Revision: 1.5 $ $Date: 2001/11/13 09:50:33 $
 */
public class ListMutationExample extends AbstractAlgorithm {
	
	// overload setup
	protected void setup() {
		
		// create initialization (sources)
		ListInitialization init1 = new ListInitialization( 2, 10 );
		ListInitialization init2 = new ListInitialization( 2, 20 );
		
		// create sinks
		GenericSink snk1 = new GenericSink();
		GenericSink snk2 = new GenericSink();
		
		// create mutation operators
		ReverseArrayMutation ram = new ReverseArrayMutation( (float) 1.0 );
		ScrambleArrayMutation sam = new ScrambleArrayMutation( (float) 1.0 );
		
		// create individual mutation operators
		IndividualMutation im1 = new IndividualMutation( new ChromosomeMutation[] { ram } );
		IndividualMutation im2 = new IndividualMutation( new ChromosomeMutation[] { sam } );
		
		// create print processors
		PrintProcessor print1pre = new PrintProcessor( "print1pre.log", false );
		PrintProcessor print1post = new PrintProcessor( "print1post.log", false );
		
		PrintProcessor print2pre = new PrintProcessor( "print2pre.log", false );
		PrintProcessor print2post = new PrintProcessor( "print2post.log", false );
		
		// connect everything
		connect( init1, print1pre );
		connect( print1pre, im1 );
		connect( im1, print1post );
		connect( print1post, snk1 );
		
		connect( init2, print2pre );
		connect( print2pre, im2 );
		connect( im2, print2post );
		connect( print2post, snk2 );
		
	}
	
	public static void main( String[] args ) throws Exception {
		ListMutationExample lme = new ListMutationExample();
		lme.start( lme, args );
	}
	
}
