
/*
 * Copyright (C) 2000-2001 Technical University of Ilmenau, Germany
 *                         Department of Microelectronic Circuits and Systems
 *
 * You may only use this file according to the terms of the license agreement included
 * within the file LICENSE which is distributed together with the software package.
 *
 * CVS Info:
 * $Author: rummler $
 * $Date: 2001/11/12 17:00:41 $
 * $Revision: 1.5 $
 * $State: Exp $
 * $RCSfile: Salesman.java,v $
 *
 */

package mss.ea.ex.salesman;

import java.util.*;

import mss.ea.al.*;
import mss.ea.core.*;
import mss.ea.eval.*;
import mss.ea.ins.*;
import mss.ea.mut.*;
import mss.ea.rec.*;
import mss.ea.sel.*;
import mss.ea.stat.*;
import mss.ea.trm.*;

public class Salesman extends AbstractAlgorithm {
	
	protected int bestKnown;
	
	// overload setup
	protected void setup() {
		
		// create initialization (sources)
		SalesmanInit init = new SalesmanInit( 100, 29 );
		
		// create comparator
		DefaultIndividualComparator comp = new DefaultIndividualComparator( new ReverseScoreComparator() );
		
		// create population
		SortedSteadyPopulation gate = new SortedSteadyPopulation( comp, 100 );
		
		// create sinks
		GenericSink snk = new GenericSink();
		
		// create processors
		// evaluation
		SalesmanCalc eval = new SalesmanCalc( "bays29" );
		
		// selection
		PercentageTruncationSelection sel = new PercentageTruncationSelection( 100, (float) 0.3 );
		
		// recombination
		ChromosomeRecombination[] cra = new ChromosomeRecombination[1];
		cra[0] = new OXArrayRecombination();
		IndividualRecombination rec = new IndividualRecombination( cra );
		
		// mutation
		ChromosomeMutation[] cma = new ChromosomeMutation[1];
		cma[0] = new ScrambleArrayMutation( (float) 1.0, (float) 0.5 );
		IndividualMutation mut = new IndividualMutation( cma );
		
		// create the condition and add them to the conditional thread
		CounterCondition cc = new CounterCondition( gate.getCounter(), 2000 );
		AbsoluteScoreStreamCondition assc = new AbsoluteScoreStreamCondition( 0, 3200 );
		assc.setComparator( comp );
		
		// create conditional router
		MultiConditionalRouter cond = new MultiConditionalRouter( cc );
		cond.addCondition( assc );
		
		DefaultIndividualCollector coll = new DefaultIndividualCollector();

		PrintProcessor print = new PrintProcessor( "print.log", false );
		PrintProcessor print2 = new PrintProcessor( "print2.log", false );
		
		// connect everything together
		connect( init, coll, 0 );
		connect( coll, eval );
		connect( eval, gate );
//		connect( eval, print );
//		connect( print, gate );
		connect( gate, cond );
//		connect( gate, print );
//		connect( print, cond );
		connect( cond, 0, sel );
		connect( cond, 1, snk );
		connect( sel, rec );
		connect( rec, mut );
		connect( mut, coll, 1 );
		
	}
	
	/*
	protected void result() {
		Log.info( "length of best found route : " + ind.getScore() + "; best known route : 2020" );
		
		Log.info( "extracting results" );
		// get the best individual
		Individual ind = (Individual) ((LinkedList) snk.getSink().getContents()).get( 0 );
		
		Log.info( "result extraction finished" );
	}
	*/
	
	public static void main( String[] args ) throws Exception {
		Salesman algorithm = new Salesman();
		algorithm.start( algorithm, args );
	}
	
}
