
/*
 * Copyright (C) 2000-2001 Technical University of Ilmenau, Germany
 *                         Department of Microelectronic Circuits and Systems
 *
 * You may only use this file according to the terms of the license agreement included
 * within the file LICENSE which is distributed together with the software package.
 *
 * CVS Info:
 * $Author: rummler $
 * $Date: 2001/11/12 17:00:42 $
 * $Revision: 1.3 $
 * $State: Exp $
 * $RCSfile: SalesmanMatrix.java,v $
 *
 */

package mss.ea.ex.salesman;

import java.io.*;
import java.util.*;

public class SalesmanMatrix {
	
	protected String comment;
	protected int[] bestRoute;
	protected int bestValue;
	protected ArrayList fileBuffer;
	protected int[][] dist;
	
	public SalesmanMatrix( String matrix ) {
		fileBuffer = new ArrayList();
		process( read( matrix ) );
	}
	
	protected ArrayList read( String file ) {
		
		LineNumberReader reader = null;
		String line = null;
		ArrayList buffer = new ArrayList();
		
		try {
			reader = new LineNumberReader( new FileReader( file ) );
		}
		catch ( FileNotFoundException e ) {
			System.err.println( "error : matrix file " + file + " not found." );
		}
		
		do {
			try {
				line = reader.readLine();
			}
			catch ( IOException io ) {
				System.err.println( "Error : read failure." );
			}
			if ( line != null ) {
				buffer.add( line );
			}
		} while ( line != null );
		return buffer;
	}
	
	protected void process( ArrayList buffer ) {
		StringTokenizer st = new StringTokenizer( (String) buffer.get( 0 ) );
		// number of cities
		int tokens = st.countTokens();
		// get the best route
		bestRoute = new int[tokens];
		for ( int i = 0; i < tokens; i++ ) {
			bestRoute[i] = Integer.parseInt( (String) st.nextToken() ) - 1;
		}
		// get best known value
		bestValue = Integer.parseInt( (String) buffer.get( 1 ) );
		// get the comment
		comment = (String) buffer.get( 2 );
		
		dist = new int[tokens][tokens];
		for ( int i = 0; i < tokens; i++ ) {
			st = new StringTokenizer( (String) buffer.get( i + 3 ) );
			int tokenNumber = 0;
			while ( st.hasMoreTokens() ) {
				int token = Integer.parseInt( st.nextToken() );
				dist[i][tokenNumber] = token;
				dist[tokenNumber][i] = token;
				tokenNumber++;
			}
		}
	}
	
	public String getComment() {
		return comment;
	}
	
	public int[] bestRoute() {
		return bestRoute;
	}
	
	public int bestValue() {
		return bestValue;
	}
	
	public int getDistance( int x, int y ) {
		return dist[x][y];
	}
	
	public int getDistance( int[] sequence ) {
		int distance = 0;
		for ( int i = 0; i < sequence.length - 1; i++ ) {
			distance += getDistance( sequence[i], sequence[i+1] );
		}
		distance += getDistance( sequence[sequence.length-1], sequence[0] ); // because of symmetry
		return distance;
	}
	
}
